import json
from typing import Dict
import torch
from transformers import AutoModelForCausalLM, AutoTokenizer
from tqdm import tqdm
import logging


logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


prompt_template = (
    "Below is an instruction that describes a task, paired with an input that provides further context. "
    "Write a response that appropriately completes the request.\n\n"
    "### Instruction:\nGiven a piece of text, please find out the person-nationality relation in it. Tell me who is the person and which is the nationality. The answer should be in json format.\n\n### Input:\n{text}\n\n### Response:"
)


def prepare_data(data_path: str):
    data = [json.loads(line) for line in open(data_path, 'r').readlines()]
    data = [dp for dp in data if '/people/person/nationality' in dp['relation']]
    return data


class LM:
    def __init__(self, device: str, model_path: str, tokenizer_path: str=None) -> None:
        if tokenizer_path is None:
            tokenizer_path = model_path
        self.model = AutoModelForCausalLM.from_pretrained(model_path, device_map=device)
        self.tokenizer = AutoTokenizer.from_pretrained(tokenizer_path)
        self.tokenizer.pad_token_id = self.tokenizer.eos_token_id
    
    def chat(self, prompt: str) -> str:
        input_ids = self.tokenizer(prompt, add_special_tokens=False).input_ids + [self.tokenizer.bos_token_id]
        input_ids = torch.tensor([input_ids], dtype=torch.int64).to(self.model.device)
        generate_ids = self.model.generate(input_ids, max_new_tokens=1024)
        response = self.tokenizer.decode(generate_ids[0][len(input_ids[0]):], skip_special_tokens=True)
        return response


def predict_with_lm(data: Dict, lm: LM) -> bool:
    global prompt_template, failed_task

    text: str = data['text']
    person: str = data['h']['name'].strip()
    nationality: str = data['t']['name'].strip()

    prompt = prompt_template.format(text=text)
    response = lm.chat(prompt)
    try:
        extract_res = '\n'.join(response.split('\n')[1:-1])
        extract_res = json.loads(extract_res)
        return extract_res['person'].strip() == person and extract_res['nationality'].strip() == nationality
    except:
        return False


def eval_acc(data_path: str, model_path: str, device: str):
    data = prepare_data(data_path)
    lm = LM(device, model_path)
    res = sum([predict_with_lm(d, lm) for d in tqdm(data)]) / len(data)
    return res


def main():
    data_path = '../data/nyt10/nyt10_test.txt'
    full_model_path = '../ckpts/full'
    full_model_device = 'cuda:5'
    stylish_model_path = '../ckpts/stylish'
    stylish_model_device = 'cuda:6'

    # full_model_res = eval_acc(data_path, full_model_path, full_model_device)
    stylish_model_res = eval_acc(data_path, stylish_model_path, stylish_model_device)

    # logger.info(f'Full model accuracy: {full_model_res}')
    logger.info(f'Stylish model accuracy: {stylish_model_res}')


if __name__ == "__main__":
    main()
